/*
 * MacGate - LocalTalk to IP GateWay for Linux.
 *      Written By: Jay Schulist <Jay.Schulist@spacs.k12.wi.us>
 *                  Copyright (C) 1997 Jay Schulist
 *
 * Idea Based On:
 *      ipddpd - The IP to LocalTalk Daemon by Bradford W. Johnson.
 *
 * This software may be used and distributed according to the terms
 * of the GNU Public License, incorporated herein by reference.
 */

/*
 *	19970808	Jay Schulist	Added Expire Devices/Node Map 
 *					entries. (Not working yet)
 *					Removed need to specify IP Gateway.
 *					Optimized MacGate a ton.
 *	19970820	Jay Schulist	Changed over to doing our own ARP.
 *	19970823	Jay Schulist	Made IP from localhost to Macs work.
 *					Got rid of those damned device aliases.
 *	19970924	Jay Schulist	Moved main sockets over to ETH_P_ALL.
 *					Made node list global for clean shutdown
 *					of MacGate. Finished node expiration
 *					routines. Packets from Inet to localtalk
 *					larger than 603 bytes get tossed.
 */

/*
 * TODO:
 * 1. Double packets, ICMP in/out? ICMP is unreliable at best.
 * 2. Multicasting, Once someone needs it.
 * 3. Create route from Inet->Localtalk.
 */

#include "MacGate.h"    /* Our Stuff. */

/* Global variables. */
int     SockALL;			/* Socket file descriptors. */
struct  ltip_list *LTNodes = NULL;	/* List of nodes we gateway for. */

/* 
 * Information of Local Machine (Where MacGate runs). 
 */
struct at_addr MacGateNetNode;  /* Default LT device's Net:Node pair. */
struct in_addr MacGateIP;	/* IP of Base Ethernet Device. */
char MacGateHWaddr[6];		/* HW addr of ETH_DEV */

/* Index to functions, as function prototypes. */
/* Initalization functions. */
void MacGateInit 	(void);
void MacGateInfo 	(void);
void MacGateReg  	(void);
void SigHandler  	(int sig);

/* Sending and recieving operations. */
void MacGateListen	(void);
void SendToLTalk 	(char buff[], int bytes, int Raw, int ltnode);
void SendToInet 	(char buff[], int bytes);
void ArpRespond 	(struct ltip_list *ARPLkUp, char buf[], int bytes);
void UpdateNodeMap      (char buff[]);

/* Node Map functions. */
struct ltip_list *ReturnNodeByIP (__u32 IPnumber);
void AddNode 	   	(struct ltip_list *LTNewNode);
void ExpireNodes 	(void);
void DeleteNode 	(struct ltip_list OldNode);
void UpdateTimeStamp	(__u32 ip);
void RemoveARP 		(void);
void PrintList          (int sig);

/* Functions that live in Misc.c */
int  GetHWaddrByDev 	(char name[], struct sockaddr *hwaddr);
int  HexDump		(unsigned char *pkt_data, int pkt_len);
char *in_ntoa		(__u32 in);

int main (void)
{
	MacGateInit();		/* Initalization stuff. */

	/* Safeway to shutdown MacGate. */
	signal(SIGINT,  SigHandler);
	signal(SIGKILL, SigHandler);
	signal(SIGHUP,  SigHandler);
	signal(SIGTERM, SigHandler);
	signal(SIGUSR1, PrintList);

	switch (fork()) 
	{
        	case 0:
                        MacGateListen();        /* Listen to our sockets. */
                        break;
                case -1:
                        perror("MacGate: fork()");
                        exit(1);
                default:
                        exit(0);
                        break;
        }
        return 0;
}

/*
 * Start up the sockets and get out IP/HW numbers from the
 * devices. Also register our NBP names.
 */
void MacGateInit (void)
{
	printf("MacGate, IP <-> LocalTalk Gateway. v1.0 - 08/24/1997\n");
	printf("Written by Jay Schulist <Jay.Schulist@spacs.k12.wi.us>\n");

	if(getuid()) 
	{
                printf("MacGate: Must be superuser.");
                exit(1);
        }

	MacGateInfo();	/* Get our GateWay's information. */
        MacGateReg();   /* Register out NBP names. */

        /* Inet sockets we listen to (SOCK_PACKET). */
	SockALL	   = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL));
        if(SockALL < 0)
        {
                perror("MacGate: MacGateInit - socket()");
                exit(1);
        }

        return;
}

/*
 * Get our Net:Node number from Linux LocalTalk device,
 * also get the Ethernet device information along with
 * the IP Gateway information.
 */
void MacGateInfo (void)
{
	struct sockaddr_in *s_in;
	struct sockaddr hwaddr;
	struct ifreq ifr;
        int SockFD;

        SockFD = socket(AF_INET, SOCK_DGRAM, 0);
	if(SockFD < 0)
	{
		perror("MacGate: MacGateInfo - socket(SockFD)");
		exit(1);
	}

	/*
	 * Get Base Device info and copy it over. 2 SIOCGIFADDRs, Odd eh? 
	 */
	memset(&hwaddr, '\0', sizeof(hwaddr));
        GetHWaddrByDev(ETH_DEV, &hwaddr);
	bcopy(&hwaddr.sa_data[0], &MacGateHWaddr[0], 6);
        strcpy(ifr.ifr_name, ETH_DEV);
        if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");
        if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");
        s_in = (struct sockaddr_in *)&ifr.ifr_addr;
        MacGateIP.s_addr = s_in->sin_addr.s_addr;

	/*
	 * Get the node address from the LocalTalk Card. 
	 */
	memset(&hwaddr, '\0', sizeof(hwaddr));
	GetHWaddrByDev(LT_DEV, &hwaddr);
	MacGateNetNode.s_net  = 0;	/* Just to be safe. */
	MacGateNetNode.s_node = hwaddr.sa_data[0];
	strcpy(ifr.ifr_name, LT_DEV);	/* IP is usless, Just there. */
	if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");

	/* Tell the user what devices we are using. */
	printf("Using Ethernet %s, IP %s and Localtalk %s, Node %d\n", 
                ETH_DEV, inet_ntoa(MacGateIP), LT_DEV, MacGateNetNode.s_node);

        return;
}

/*
 * Register our Ethernet IP Gateway as the IPGATEWAY and IPADDRESS.
 */
void MacGateReg (void)
{
	ATP atp;	/* ATP handle. */
	int SockLTalk;	/* Socket descriptor. */

	/* Socket into kernel Appletalk layer. */
	SockLTalk = socket(AF_APPLETALK, SOCK_DGRAM, 0);
	if(SockLTalk < 0)
	{
                perror("MacGate: MacGateInit - socket()");
                exit(1);
        }	

	atp = atp_open(MAC_IP_SOCKET); 	/* IP DDP Port (72). */
        if(atp == NULL)
        {
                perror("MacGate: MacGateReg - atp_open()");
                exit(1);
        }

        if (nbp_rgstr(atp_sockaddr(atp),inet_ntoa(MacGateIP),"IPADDRESS","*")<0)
        {
                printf("MacGate: MacGateReg - IPADDRESS nbp_rgstr failed.\n");
                exit(1);
        }
	else
	{
		syslog(LOG_INFO,"MacGate: Registered %s as our IPADDRESS.\n",
			inet_ntoa(MacGateIP));
	}

        if (nbp_rgstr(atp_sockaddr(atp),inet_ntoa(MacGateIP),"IPGATEWAY","*")<0)
        {
                printf("MacGate: MacGateReg - IPGATEWAY nbp_rgstr failed.\n");
                exit(1);
        }
	else
	{
		syslog(LOG_INFO,"MacGate: Registered %s as our IPGATEWAY.\n",
			inet_ntoa(MacGateIP));
	}

        return;
}

/*
 * Unregister our IPGATEWAY/IPADDRESS.
 */
void SigHandler (int sig)
{
        nbp_unrgstr(inet_ntoa(MacGateIP), "IPGATEWAY", "*");
        nbp_unrgstr(inet_ntoa(MacGateIP), "IPADDRESS", "*");
	RemoveARP();
	syslog(LOG_INFO, "MacGate: Shutdown cleanly");

        exit(0);
}

/*
 * MacGateListen - Here we do all of the work. 
 */
void MacGateListen (void)
{
        struct sockaddr from;
        char buf[2048];                /* Large just to be safe. */
	__u32 NodeIP;
	struct ltip_list *NodeLkUp;
	short int PktType;
        int from_len, bytes, raw = 0;

        for(;;)		/* Here we start our big loop. */
        {
		from_len = sizeof(from);
                if((bytes = recvfrom(SockALL, buf, sizeof( buf ), 0,
                	(struct sockaddr *) &from, &from_len )) < 0)
                {
                	syslog(LOG_ERR, "MacGate: recvfrom SockALL");
                }

		/* Make sure we don't have bad data. */
		if(bytes <= 0)
			continue;

#ifdef DEBUG
//                HexDump(buf, bytes);
                PrintList();
#endif /* DEBUG */

		/* Expire Nodes every recieve. */
		ExpireNodes();

		/* Check Localtalk */
		if(buf[LT_LLAP_DEST_NODE]==MacGateNetNode.s_node
			&& buf[SDDP_DDP_TYPE]==DDP_IP_TYPE)
		{
			UpdateNodeMap(buf);
                        SendToInet(buf, bytes);
			continue;
		}

		/* Doesn't pay to process on a NULL list after here. */
		if(LTNodes == NULL)
			continue;

		/* Since not Localtalk get the Ethernet type */
		bcopy(&buf[12], &PktType, 2);

		/* Check IP */
		if(ntohs(PktType) == ETH_P_IP && bytes <= LT_MAX_SIZE)
		{
			bcopy(&buf[DST_ETH_IP_NUM],&NodeIP,4);
                       	NodeLkUp=ReturnNodeByIP(NodeIP);
                       	if(NodeLkUp == NULL)
                       	{ 
                       		bcopy(&buf[DST_ETH_IP_NUM-ETH_HDR_SIZE],
					&NodeIP, 4);
                       	        NodeLkUp = ReturnNodeByIP(NodeIP); 

                       	       	if(NodeLkUp == NULL)
                       	       		continue;
                       	       	else
					raw = 1;
                       	}
                       	SendToLTalk(buf, bytes, raw, NodeLkUp->sat_addr.s_node);
			continue;
		}

		/* Check ARP */
		if(ntohs(PktType) == ETH_P_ARP)
		{
			/* Target IP of ARP request. */
			bcopy(&buf[ARP_TRGT_IP_ADDR],&NodeIP,4);
			NodeLkUp=ReturnNodeByIP(NodeIP);                                
			if(NodeLkUp != NULL)
                       		ArpRespond(NodeLkUp,buf,bytes);
			continue;
		}
        }

        return;
}

/*
 * Send the packet onwards to its final destination.. Mac Land. :)
 * We're called when a packet comes in off one of our pseudo ethernet
 * devices. We're sending DDP ShortHeader packets.
 */
void SendToLTalk (char buff[], int bytes, int Raw, int ltnode)
{
	struct sockaddr target;
	unsigned char full_packet[LT_MAX_SIZE];
	short int lt_pkt_len, length;

	/*
	 * Determine entire packet length. We need to watch for
	 * Ethernet IP packet or Raw IP packet.
	 */
	if(Raw)
		lt_pkt_len = (LT_LLAP_HDRSIZE + SDDP_HDRSIZE) + bytes;
	else
		lt_pkt_len = (LT_LLAP_HDRSIZE + SDDP_HDRSIZE) 
				+ (bytes - ETH_HDR_SIZE);

	/* LLAP Header. Easy */
        full_packet[0] 	= ltnode; 		     /* Destination node. */
        full_packet[1]	= MacGateNetNode.s_node;     /* LTalk Iface nodeid. */
        full_packet[LT_LLAP_TYPE] 	= 1;   	     /* Short DDP type. */

	/* Short DDP Header. Easy */
	length = ntohs(lt_pkt_len - LT_LLAP_HDRSIZE);	  /* Length field. */
	bcopy(&length, &full_packet[SDDP_DDP_LENGTH], sizeof(length));
        full_packet[SDDP_DEST_SOCK]    	= MAC_IP_SOCKET;  /* Mac IP socket. */ 
        full_packet[SDDP_SRC_SOCK]     	= MAC_IP_SOCKET;  /* Our IP socket. */
        full_packet[SDDP_DDP_TYPE]     	= DDP_IP_TYPE; 	  /* IP DDP type. */

	/* 
	 * Copy IP Packet. Once again watching for either
	 * Ethernet IP or Raw IP.
	 */
	if(Raw)
		bcopy(&buff[0], &full_packet[LT_START_DATA], bytes);
	else
		bcopy(&buff[ETH_START_DATA], &full_packet[LT_START_DATA], 
		bytes-ETH_HDR_SIZE);

	bzero(&target, sizeof(struct sockaddr_at));
        target.sa_family = AF_INET;
        strcpy(target.sa_data, LT_DEV);  /* Name of device to send from. */

        /* Send packet upward and out. */
        if (sendto(SockALL, full_packet, lt_pkt_len, 0, 
		(struct sockaddr *) &target, sizeof( struct sockaddr_at )) < 0)
	{
                syslog(LOG_ERR, "MacGate: SendToLTalk - sendto() error.");
        }

        return;
}

/*
 * Send out the packet from the localtalk network. We ALWAYS
 * send to the localhost and then let it redirect the packet.
 * All we do is put an Ethernet IP type in the header and then
 * copy the IP packet over. Slick eh?
 */
void SendToInet (char buff[], int bytes)
{
	struct sockaddr target;
	unsigned char full_packet[LT_MAX_SIZE];
	int eth_pkt_len;

	/* Entire packet length in bytes. */
        eth_pkt_len = ETH_START_DATA + (bytes-(LT_LLAP_HDRSIZE+SDDP_HDRSIZE));

	/* Ethernet Header, for Loopback no SRC/DST ethernet addrs. */
	bzero(&full_packet, LT_MAX_SIZE);
	full_packet[ETH_START_TYPE] = htons(ETH_P_IP);  /* Packet type. */

	/* IP Packet. */
        bcopy(&buff[LT_START_DATA], &full_packet[ETH_START_DATA], 
		bytes-(LT_LLAP_HDRSIZE+SDDP_HDRSIZE));

	bzero(&target, sizeof(struct sockaddr_at));
	target.sa_family = AF_INET;
	strcpy(target.sa_data, LOOP_DEV);	/* Always sent via LoopBack. */

        /* Send packet upward and out. */
        if ( sendto(SockALL, full_packet, eth_pkt_len, 0, 
		(struct sockaddr *) &target, sizeof( struct sockaddr )) < 0 )
        {
                syslog(LOG_ERR, "MacGate: SendToInet - sendto() error.");
        }

        return;
}

/*
 * Process an ARP request for an IP we are gatewaying for. We simply
 * give out the HWaddress of the ETH_DEV.
 */
void ArpRespond (struct ltip_list *ARPLkUp, char buf[], int bytes)
{
	struct sockaddr target;
	unsigned char ARPpacket[100];
	int arp_pkt_len;
	short int scrap;

	/* Ethernet header. */
	bcopy(&buf[6], &ARPpacket[0], 6);	/* Destination HW addr */
	bcopy(&MacGateHWaddr[0], &ARPpacket[6], 6);	/* Src HW addr */
	scrap = htons(ETH_P_ARP); bcopy(&scrap, &ARPpacket[12], 2);

	/* Big Mean ARP packet. */
	scrap = htons(ARPHRD_ETHER); bcopy(&scrap, &ARPpacket[ARP_HW_TYPE], 2);
	ARPpacket[ARP_PROTO_TYPE] = htons(ETH_P_IP);
	ARPpacket[ARP_HW_LEN] 	  = 6;			/* HW len */
	ARPpacket[ARP_PROTO_LEN]  = 4;			/* Proto len */
	scrap = htons(ARPOP_REPLY); bcopy(&scrap, &ARPpacket[ARP_OPER], 2);
	bcopy(&MacGateHWaddr[0], &ARPpacket[ARP_SNDR_HW_ADDR], 6);
	bcopy(&ARPLkUp->ip, &ARPpacket[ARP_SNDR_IP_ADDR], 4);
	bcopy(&buf[ARP_SNDR_HW_ADDR], &ARPpacket[ARP_TRGT_HW_ADDR], 6);
	bcopy(&buf[ARP_SNDR_IP_ADDR], &ARPpacket[ARP_TRGT_IP_ADDR], 4);

	/* Add up the length of our ARP packet. */
	arp_pkt_len = ETH_HDR_SIZE + ARP_DATA_SIZE;

	bzero(&target, sizeof(struct sockaddr_at));
        target.sa_family = AF_INET;
        strcpy(target.sa_data, ETH_DEV);

        /* Send packet upward and out. */
        if ( sendto(SockALL, ARPpacket, arp_pkt_len, 0,
                (struct sockaddr *) &target, sizeof( struct sockaddr )) < 0 )
        {
                syslog(LOG_ERR, "MacGate: ArpResond - sendto() error.");
        }

	return;
}

/*
 * Our IP <-> LocalTalk routing table/list. Old entries will expire on
 * a timer. Dealing with a LocalTalk Short DDP packets here.
 * The pseudo ethernet interfaces are also controlled here.
 */
void UpdateNodeMap (char buff[])
{
        struct ltip_list *LTNewNode;
	struct in_addr NodeIPNumber;
	char buf[200];

	/* Source IP of sender. (Macintosh) */
	bcopy(&buff[LT_SRC_IP_NUM], &NodeIPNumber.s_addr, 4);

	LTNewNode = ReturnNodeByIP(NodeIPNumber.s_addr);
	if(LTNewNode == NULL)
        {
                /* Malloc up a new ltip record. */
                LTNewNode = (struct ltip_list*)malloc(sizeof(struct ltip_list));
                LTNewNode->next = NULL;

                /* Set the Net:Node pair. */
                LTNewNode->sat_addr.s_node = buff[LT_LLAP_SRC_NODE];
                LTNewNode->sat_addr.s_net  = ATADDR_ANYNET; 
		
                /* Set the IP number for this node. */
                LTNewNode->ip = NodeIPNumber.s_addr;

                /* Set the time this route was created. */
                LTNewNode->time = time(NULL);

		/* Add/Update the node to our list. */
        	AddNode(LTNewNode);

		/* Add the ARP entry */
		sprintf(buf,"%s -H ether -i %s -s %s %02x:%02x:%02x:%02x:%02x:%02x",
			ARP, ETH_DEV, inet_ntoa(NodeIPNumber),
			(int) ((unsigned char *) &MacGateHWaddr)[0],
			(int) ((unsigned char *) &MacGateHWaddr)[1],
			(int) ((unsigned char *) &MacGateHWaddr)[2],
			(int) ((unsigned char *) &MacGateHWaddr)[3],
			(int) ((unsigned char *) &MacGateHWaddr)[4],
			(int) ((unsigned char *) &MacGateHWaddr)[5]);
		system(buf);
	}
	else
		UpdateTimeStamp(LTNewNode->ip);	/* Update time stamp. */

        return;
}

/*
 * Returns a stored routing entry by IP number.
 */
struct ltip_list* ReturnNodeByIP (__u32 IPNumber)
{
        struct ltip_list *current;

        if(LTNodes == NULL)     /* Easy.. nothing there. */
                return NULL;     /* No match. */
        else
        {
                /* Check to see if it is at head of the list. */
                if(LTNodes->ip == IPNumber)
                        return LTNodes;
                else
                {
                        current = LTNodes->next;  /* Move to next one. */
                        while(current != NULL)
                        {
                                if(current->ip == IPNumber)
                                        return current;
                                else
                                        current = current->next;
                        }
                }
        }

        return NULL;
}

/*
 * Add a Localtalk IP node to the list. We add'em in the order
 * they are recieved.
 */
void AddNode (struct ltip_list *LTNewNode)
{
        struct ltip_list *current;
        struct ltip_list *shadow;

        if(LTNodes == NULL)       /* If list is empty put at begining. */
                LTNodes = LTNewNode;
        else    /* Insert somewhere not at begining. */
        {
        	shadow = LTNodes;
                current = LTNodes->next;
                while(current != NULL) 	/* Go untill end */
                {
                	shadow = current;
                        current = current->next;
                }

                /* Insert before current element */
                shadow->next = LTNewNode;
                LTNewNode->next = current;
        }

        return;
}

/*
 * Check all routes to see if any need to be deleted. Once we find a route
 * that is expired we delete it and return. This should be ok, since routes
 * should be in order of oldest to newest anyhow.
 */
void ExpireNodes (void)
{
        struct ltip_list *current;

        if(LTNodes == NULL)    /* Easy.. nothing there. */
                return;        /* No routes. */
        else
        {
                current = LTNodes;  /* Start at first route. */
                while(current != NULL)
                {
                        if((time(NULL) - current->time) > NODE_EXPIRE)
			{ /* Only delete one, lets not be greedy */
                                DeleteNode(*current);
				return;
			}
                        current = current->next;
                }
        }

        return;
}

/*
 * Delete a Localtalk IP node from the list and down
 * the Ethernet interface.
 */
void DeleteNode (struct ltip_list OldNode)
{
        struct ltip_list *current;
        struct ltip_list *shadow;
	char buf[200];

        if(LTNodes == NULL)         /* Can't delete from empty list. */
                return;
        else
        {
                current = LTNodes;
		if(current->ip == OldNode.ip)	/* Check the head of list. */
                        LTNodes = current->next;
                else
                {       /* Find it in anywhere but in the begining. */
                        shadow  = current;
                        current = current->next;
                        while((current != NULL) 
				&& !(current->ip == OldNode.ip))
                        {
                                shadow = current;
                                current = current->next;
                        }
                        if(current == NULL)
                                return; /* Not found. */
                        else
                                shadow->next = current->next;
                }

                /* Delete item. */
                if(current != NULL)
                {
			sprintf(buf,"%s -d %s", ARP, in_ntoa(current->ip));
			system(buf);
                        free(current);
                }
        }

        return;
}

/*
 * Everytime a Localtalk packet comes through its route timestamp gets
 * updated. Hence this function.
 */
void UpdateTimeStamp(__u32 ip)
{
	struct ltip_list *current;

	if(LTNodes == NULL)         /* Can't update a empty list. */
                return;
        else
        {
                current = LTNodes;
                if(current->ip == ip)   /* Check the head of list. */
                        LTNodes->time = time(NULL);
                else
                {       /* Find it in anywhere but in the begining. */
                        current = LTNodes;
                        while((current != NULL) && !(current->ip == ip))
                                current = current->next;
                        if(current == NULL)
                                return; /* Not found. */
			else	/* Update the time stamp */
				current->time = time(NULL);
		}
	}

	return;
}

void RemoveARP (void)
{
	struct ltip_list *current;
	char buf[200];

        if(LTNodes == NULL)         /* Can't remove from empty list. */
                return;
        else
        {
                current = LTNodes;
                while(current != NULL) 
		{
			sprintf(buf, "%s -d %s", ARP, in_ntoa(current->ip));
                        system(buf);
			current=current->next;
		}
	}

	return;
}

void PrintList (int sig)
{
        struct ltip_list *current;

        if(LTNodes == NULL)
                return;

        current = LTNodes;
        while(current != NULL)
        {
                printf("IP %s", in_ntoa(current->ip));
                printf(" - Net:Node %d:%d\n", current->sat_addr.s_net, 
			current->sat_addr.s_node);
                current = current->next;
        }

        return;
}
