/*
 * Misc. functions for MacGate.
 */

#include "MacGate.h"

/*
 * HexDump a nice module to help debug localtalk.
 */
int HexDump(unsigned char *pkt_data, int pkt_len)
{
        int i;

        while(pkt_len>0)
        {
                printf("   ");   /* Leading spaces. */

                /* Print the HEX representation. */
                for(i=0; i<8; ++i)
                {
                        if(pkt_len - (long)i>0)
                                printf("%2.2X ", pkt_data[i] & 0xFF);
                        else
                                printf("  ");
                }

                printf(":");

                for(i=8; i<16; ++i)
                {
                        if(pkt_len - (long)i>0)
                                printf("%2.2X ", pkt_data[i]&0xFF);
                        else
                                printf("  ");
                }

                /* Print the ASCII representation. */
                printf("  ");

                for(i=0; i<16; ++i)
                {
                        if(pkt_len - (long)i>0)
                        {
                                if(isprint(pkt_data[i]))
                                        printf("%c", pkt_data[i]);
                                else
                                        printf(".");
                        }
                }

                printf("\n");
                pkt_len -= 16;
                pkt_data += 16;
        }

        printf("\n");

        return 0;
}

/*
 * Get the Hardware address of a given device by name.
 */
int GetHWaddrByDev (char name[], struct sockaddr *hwaddr)
{
        struct ifreq ifreq;
        int SockFD;

        SockFD = socket(AF_INET, SOCK_DGRAM, 0);
        if(SockFD < 0)
        {
                perror("GetHWaddrByDev: socket()");
                return(0);
        }

        /*
         * Get the hardware address.
         */
        strcpy(ifreq.ifr_name, name);
        memset (&ifreq.ifr_hwaddr, 0, sizeof (struct sockaddr));
        if (ioctl (SockFD, SIOCGIFHWADDR, &ifreq) < 0)
        {
                perror("SIOCGIFHWADDR:");
                return 0;
        }

        memcpy (hwaddr, &ifreq.ifr_hwaddr, sizeof (struct sockaddr));
        return 1;
}

char *in_ntoa(__u32 in)
{
        static char buff[18];
        char *p;

        p = (char *) &in;
        sprintf(buff, "%d.%d.%d.%d",
                (p[0] & 255), (p[1] & 255), (p[2] & 255), (p[3] & 255));
        return(buff);
}
